@@ -2,23 +2,26 @@
  * Copyright © 2008 Kristian Høgsberg
  * Copyright © 2013 Jason Ekstrand
  *
- * Permission to use, copy, modify, distribute, and sell this software and its
- * documentation for any purpose is hereby granted without fee, provided that
- * the above copyright notice appear in all copies and that both that copyright
- * notice and this permission notice appear in supporting documentation, and
- * that the name of the copyright holders not be used in advertising or
- * publicity pertaining to distribution of the software without specific,
- * written prior permission.  The copyright holders make no representations
- * about the suitability of this software for any purpose.  It is provided "as
- * is" without express or implied warranty.
+ * Permission is hereby granted, free of charge, to any person obtaining
+ * a copy of this software and associated documentation files (the
+ * "Software"), to deal in the Software without restriction, including
+ * without limitation the rights to use, copy, modify, merge, publish,
+ * distribute, sublicense, and/or sell copies of the Software, and to
+ * permit persons to whom the Software is furnished to do so, subject to
+ * the following conditions:
  *
- * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
- * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
- * EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL, INDIRECT OR
- * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
- * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
- * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
- * OF THIS SOFTWARE.
+ * The above copyright notice and this permission notice (including the
+ * next paragraph) shall be included in all copies or substantial
+ * portions of the Software.
+ *
+ * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
+ * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
+ * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
+ * NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
+ * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
+ * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
+ * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
+ * SOFTWARE.
  */
 
 #define _GNU_SOURCE
@@ -30,7 +33,6 @@
 #include <stdio.h>
 #include <errno.h>
 #include <sys/uio.h>
-#include <assert.h>
 #include <fcntl.h>
 #include <unistd.h>
 #include <sys/types.h>
@@ -61,11 +63,18 @@ struct wl_connection {
 	int want_flush;
 };
 
-static void
+static int
 wl_buffer_put(struct wl_buffer *b, const void *data, size_t count)
 {
 	uint32_t head, size;
 
+	if (count > sizeof(b->data)) {
+		wl_log("Data too big for buffer (%d > %d).\n",
+		       count, sizeof(b->data));
+		errno = E2BIG;
+		return -1;
+	}
+
 	head = MASK(b->head);
 	if (head + count <= sizeof b->data) {
 		memcpy(b->data + head, data, count);
@@ -76,6 +85,8 @@ wl_buffer_put(struct wl_buffer *b, const void *data, size_t count)
 	}
 
 	b->head += count;
+
+	return 0;
 }
 
 static void
@@ -152,10 +163,10 @@ wl_connection_create(int fd)
 {
 	struct wl_connection *connection;
 
-	connection = malloc(sizeof *connection);
+	connection = zalloc(sizeof *connection);
 	if (connection == NULL)
 		return NULL;
-	memset(connection, 0, sizeof *connection);
+
 	connection->fd = fd;
 
 	return connection;
@@ -180,13 +191,16 @@ close_fds(struct wl_buffer *buffer, int max)
 	buffer->tail += size;
 }
 
-void
+int
 wl_connection_destroy(struct wl_connection *connection)
 {
+	int fd = connection->fd;
+
 	close_fds(&connection->fds_out, -1);
 	close_fds(&connection->fds_in, -1);
-	close(connection->fd);
 	free(connection);
+
+	return fd;
 }
 
 void
@@ -243,8 +257,8 @@ decode_cmsg(struct wl_buffer *buffer, struct msghdr *msg)
 			size /= sizeof(int32_t);
 			for (i = 0; i < size; i++)
 				close(((int*)CMSG_DATA(cmsg))[i]);
-		} else {
-			wl_buffer_put(buffer, CMSG_DATA(cmsg), size);
+		} else if (wl_buffer_put(buffer, CMSG_DATA(cmsg), size) < 0) {
+				return -1;
 		}
 	}
 
@@ -278,7 +292,7 @@ wl_connection_flush(struct wl_connection *connection)
 		msg.msg_namelen = 0;
 		msg.msg_iov = iov;
 		msg.msg_iovlen = count;
-		msg.msg_control = cmsg;
+		msg.msg_control = (clen > 0) ? cmsg : NULL;
 		msg.msg_controllen = clen;
 		msg.msg_flags = 0;
 
@@ -300,6 +314,12 @@ wl_connection_flush(struct wl_connection *connection)
 	return connection->out.head - tail;
 }
 
+uint32_t
+wl_connection_pending_input(struct wl_connection *connection)
+{
+	return wl_buffer_size(&connection->in);
+}
+
 int
 wl_connection_read(struct wl_connection *connection)
 {
@@ -336,7 +356,7 @@ wl_connection_read(struct wl_connection *connection)
 
 	connection->in.head += len;
 
-	return connection->in.head - connection->in.tail;
+	return wl_connection_pending_input(connection);
 }
 
 int
@@ -350,7 +370,9 @@ wl_connection_write(struct wl_connection *connection,
 			return -1;
 	}
 
-	wl_buffer_put(&connection->out, data, count);
+	if (wl_buffer_put(&connection->out, data, count) < 0)
+		return -1;
+
 	connection->want_flush = 1;
 
 	return 0;
@@ -367,9 +389,7 @@ wl_connection_queue(struct wl_connection *connection,
 			return -1;
 	}
 
-	wl_buffer_put(&connection->out, data, count);
-
-	return 0;
+	return wl_buffer_put(&connection->out, data, count);
 }
 
 static int
@@ -385,6 +405,12 @@ wl_message_count_arrays(const struct wl_message *message)
 	return arrays;
 }
 
+int
+wl_connection_get_fd(struct wl_connection *connection)
+{
+	return connection->fd;
+}
+
 static int
 wl_connection_put_fd(struct wl_connection *connection, int32_t fd)
 {
@@ -394,9 +420,7 @@ wl_connection_put_fd(struct wl_connection *connection, int32_t fd)
 			return -1;
 	}
 
-	wl_buffer_put(&connection->fds_out, &fd, sizeof fd);
-
-	return 0;
+	return wl_buffer_put(&connection->fds_out, &fd, sizeof fd);
 }
 
 const char *
@@ -512,7 +536,7 @@ wl_closure_marshal(struct wl_object *sender, uint32_t opcode,
 
 	count = arg_count_for_signature(message->signature);
 	if (count > WL_CLOSURE_MAX_ARGS) {
-		printf("too many args (%d)\n", count);
+		wl_log("too many args (%d)\n", count);
 		errno = EINVAL;
 		return NULL;
 	}
@@ -556,16 +580,12 @@ wl_closure_marshal(struct wl_object *sender, uint32_t opcode,
 		case 'h':
 			fd = args[i].h;
 			dup_fd = wl_os_dupfd_cloexec(fd, 0);
-			if (dup_fd < 0) {
-				fprintf(stderr, "dup failed: %m");
-				abort();
-			}
+			if (dup_fd < 0)
+				wl_abort("dup failed: %s\n", strerror(errno));
 			closure->args[i].h = dup_fd;
 			break;
 		default:
-			fprintf(stderr, "unhandled format code: '%c'\n",
-				arg.type);
-			assert(0);
+			wl_abort("unhandled format code: '%c'\n", arg.type);
 			break;
 		}
 	}
@@ -615,7 +635,7 @@ wl_connection_demarshal(struct wl_connection *connection,
 
 	count = arg_count_for_signature(message->signature);
 	if (count > WL_CLOSURE_MAX_ARGS) {
-		printf("too many args (%d)\n", count);
+		wl_log("too many args (%d)\n", count);
 		errno = EINVAL;
 		wl_connection_consume(connection, size);
 		return NULL;
@@ -642,7 +662,7 @@ wl_connection_demarshal(struct wl_connection *connection,
 		signature = get_next_argument(signature, &arg);
 
 		if (arg.type != 'h' && p + 1 > end) {
-			printf("message too short, "
+			wl_log("message too short, "
 			       "object (%d), message %s(%s)\n",
 			       *p, message->name, message->signature);
 			errno = EINVAL;
@@ -669,7 +689,7 @@ wl_connection_demarshal(struct wl_connection *connection,
 
 			next = p + DIV_ROUNDUP(length, sizeof *p);
 			if (next > end) {
-				printf("message too short, "
+				wl_log("message too short, "
 				       "object (%d), message %s(%s)\n",
 				       closure->sender_id, message->name,
 				       message->signature);
@@ -680,7 +700,7 @@ wl_connection_demarshal(struct wl_connection *connection,
 			s = (char *) p;
 
 			if (length > 0 && s[length - 1] != '\0') {
-				printf("string not nul-terminated, "
+				wl_log("string not nul-terminated, "
 				       "message %s(%s)\n",
 				       message->name, message->signature);
 				errno = EINVAL;
@@ -695,7 +715,7 @@ wl_connection_demarshal(struct wl_connection *connection,
 			closure->args[i].n = id;
 
 			if (id == 0 && !arg.nullable) {
-				printf("NULL object received on non-nullable "
+				wl_log("NULL object received on non-nullable "
 				       "type, message %s(%s)\n", message->name,
 				       message->signature);
 				errno = EINVAL;
@@ -707,7 +727,7 @@ wl_connection_demarshal(struct wl_connection *connection,
 			closure->args[i].n = id;
 
 			if (id == 0 && !arg.nullable) {
-				printf("NULL new ID received on non-nullable "
+				wl_log("NULL new ID received on non-nullable "
 				       "type, message %s(%s)\n", message->name,
 				       message->signature);
 				errno = EINVAL;
@@ -715,7 +735,7 @@ wl_connection_demarshal(struct wl_connection *connection,
 			}
 
 			if (wl_map_reserve_new(objects, id) < 0) {
-				printf("not a valid new object id (%d), "
+				wl_log("not a valid new object id (%u), "
 				       "message %s(%s)\n",
 				       id, message->name, message->signature);
 				errno = EINVAL;
@@ -728,7 +748,7 @@ wl_connection_demarshal(struct wl_connection *connection,
 
 			next = p + DIV_ROUNDUP(length, sizeof *p);
 			if (next > end) {
-				printf("message too short, "
+				wl_log("message too short, "
 				       "object (%d), message %s(%s)\n",
 				       closure->sender_id, message->name,
 				       message->signature);
@@ -745,7 +765,7 @@ wl_connection_demarshal(struct wl_connection *connection,
 			break;
 		case 'h':
 			if (connection->fds_in.tail == connection->fds_in.head) {
-				printf("file descriptor expected, "
+				wl_log("file descriptor expected, "
 				       "object (%d), message %s(%s)\n",
 				       closure->sender_id, message->name,
 				       message->signature);
@@ -758,8 +778,7 @@ wl_connection_demarshal(struct wl_connection *connection,
 			closure->args[i].h = fd;
 			break;
 		default:
-			printf("unknown type\n");
-			assert(0);
+			wl_abort("unknown type\n");
 			break;
 		}
 	}
@@ -817,7 +836,7 @@ wl_closure_lookup_objects(struct wl_closure *closure, struct wl_map *objects)
 				 * destroyed client side */
 				object = NULL;
 			} else if (object == NULL && id != 0) {
-				printf("unknown object (%u), message %s(%s)\n",
+				wl_log("unknown object (%u), message %s(%s)\n",
 				       id, message->name, message->signature);
 				object = NULL;
 				errno = EINVAL;
@@ -827,7 +846,7 @@ wl_closure_lookup_objects(struct wl_closure *closure, struct wl_map *objects)
 			if (object != NULL && message->types[i] != NULL &&
 			    !wl_interface_equal((object)->interface,
 						message->types[i])) {
-				printf("invalid object (%u), type (%s), "
+				wl_log("invalid object (%u), type (%s), "
 				       "message %s(%s)\n",
 				       id, (object)->interface->name,
 				       message->name, message->signature);
@@ -893,8 +912,7 @@ convert_arguments_to_ffi(const char *signature, uint32_t flags,
 			ffi_args[i] = &args[i].h;
 			break;
 		default:
-			printf("unknown type\n");
-			assert(0);
+			wl_abort("unknown type\n");
 			break;
 		}
 	}
@@ -924,6 +942,10 @@ wl_closure_invoke(struct wl_closure *closure, uint32_t flags,
 		     count + 2, &ffi_type_void, ffi_types);
 
 	implementation = target->implementation;
+	if (!implementation[opcode]) {
+		wl_abort("listener function for opcode %u of %s is NULL\n",
+			 opcode, target->interface->name);
+	}
 	ffi_call(&cif, implementation[opcode], NULL, ffi_args);
 }
 
@@ -953,8 +975,8 @@ copy_fds_to_connection(struct wl_closure *closure,
 
 		fd = closure->args[i].h;
 		if (wl_connection_put_fd(connection, fd)) {
-			fprintf(stderr, "request could not be marshaled: "
-				"can't send file descriptor");
+			wl_log("request could not be marshaled: "
+			       "can't send file descriptor");
 			return -1;
 		}
 	}
@@ -1115,7 +1137,7 @@ wl_closure_send(struct wl_closure *closure, struct wl_connection *connection)
 		return -1;
 
 	buffer_size = buffer_size_for_closure(closure);
-	buffer = malloc(buffer_size * sizeof buffer[0]);
+	buffer = zalloc(buffer_size * sizeof buffer[0]);
 	if (buffer == NULL)
 		return -1;
 
