@@ -1,46 +1,126 @@
 /*
  * Copyright © 2008-2011 Kristian Høgsberg
  * Copyright © 2011 Intel Corporation
+ * Copyright © 2015 Red Hat, Inc.
  *
- * Permission to use, copy, modify, distribute, and sell this software and its
- * documentation for any purpose is hereby granted without fee, provided that
- * the above copyright notice appear in all copies and that both that copyright
- * notice and this permission notice appear in supporting documentation, and
- * that the name of the copyright holders not be used in advertising or
- * publicity pertaining to distribution of the software without specific,
- * written prior permission.  The copyright holders make no representations
- * about the suitability of this software for any purpose.  It is provided "as
- * is" without express or implied warranty.
+ * Permission is hereby granted, free of charge, to any person obtaining
+ * a copy of this software and associated documentation files (the
+ * "Software"), to deal in the Software without restriction, including
+ * without limitation the rights to use, copy, modify, merge, publish,
+ * distribute, sublicense, and/or sell copies of the Software, and to
+ * permit persons to whom the Software is furnished to do so, subject to
+ * the following conditions:
  *
- * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
- * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
- * EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL, INDIRECT OR
- * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
- * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
- * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
- * OF THIS SOFTWARE.
+ * The above copyright notice and this permission notice (including the
+ * next paragraph) shall be included in all copies or substantial
+ * portions of the Software.
+ *
+ * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
+ * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
+ * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
+ * NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
+ * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
+ * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
+ * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
+ * SOFTWARE.
  */
 
+#include "config.h"
+
+#include <stdbool.h>
 #include <stdio.h>
 #include <stdarg.h>
 #include <string.h>
 #include <errno.h>
 #include <ctype.h>
 #include <expat.h>
+#include <getopt.h>
+#include <limits.h>
+#include <unistd.h>
+
+#if HAVE_LIBXML
+#include <libxml/parser.h>
+#endif
 
 #include "wayland-util.h"
 
+/* Embedded wayland.dtd file, see dtddata.S */
+extern char DTD_DATA_begin;
+extern int DTD_DATA_len;
+
+enum side {
+	CLIENT,
+	SERVER,
+};
+
 static int
 usage(int ret)
 {
-	fprintf(stderr, "usage: ./scanner [client-header|server-header|code]\n");
+	fprintf(stderr, "usage: ./scanner [OPTION] [client-header|server-header|code]"
+		" [input_file output_file]\n");
 	fprintf(stderr, "\n");
 	fprintf(stderr, "Converts XML protocol descriptions supplied on "
-			"stdin to client headers,\n"
-			"server headers, or protocol marshalling code.\n");
+			"stdin or input file to client\n"
+			"headers, server headers, or protocol marshalling code.\n\n");
+	fprintf(stderr, "options:\n");
+	fprintf(stderr, "    -h,  --help                  display this help and exit.\n"
+		        "    -c,  --include-core-only     include the core version of the headers,\n"
+	                "                                 that is e.g. wayland-client-core.h instead\n"
+	                "                                 of wayland-client.h.\n");
 	exit(ret);
 }
 
+static bool
+is_dtd_valid(FILE *input, const char *filename)
+{
+	bool rc = true;
+#if HAVE_LIBXML
+	xmlParserCtxtPtr ctx = NULL;
+	xmlDocPtr doc = NULL;
+	xmlDtdPtr dtd = NULL;
+	xmlValidCtxtPtr	dtdctx;
+	xmlParserInputBufferPtr	buffer;
+	int fd = fileno(input);
+
+	dtdctx = xmlNewValidCtxt();
+	ctx = xmlNewParserCtxt();
+	if (!ctx || !dtdctx)
+		abort();
+
+	buffer = xmlParserInputBufferCreateMem(&DTD_DATA_begin,
+					       DTD_DATA_len,
+					       XML_CHAR_ENCODING_UTF8);
+	if (!buffer) {
+		fprintf(stderr, "Failed to init buffer for DTD.\n");
+		abort();
+	}
+
+	dtd = xmlIOParseDTD(NULL, buffer, XML_CHAR_ENCODING_UTF8);
+	if (!dtd) {
+		fprintf(stderr, "Failed to parse DTD.\n");
+		abort();
+	}
+
+	doc = xmlCtxtReadFd(ctx, fd, filename, NULL, 0);
+	if (!doc) {
+		fprintf(stderr, "Failed to read XML\n");
+		abort();
+	}
+
+	rc = xmlValidateDtd(dtdctx, doc, dtd);
+	xmlFreeDoc(doc);
+	xmlFreeParserCtxt(ctx);
+	xmlFreeValidCtxt(dtdctx);
+	/* xmlIOParseDTD consumes buffer */
+
+	if (lseek(fd, 0, SEEK_SET) != 0) {
+		fprintf(stderr, "Failed to reset fd, output would be garbage.\n");
+		abort();
+	}
+#endif
+	return rc;
+}
+
 #define XML_BUFFER_SIZE 4096
 
 struct location {
@@ -61,6 +141,7 @@ struct protocol {
 	int null_run_length;
 	char *copyright;
 	struct description *description;
+	bool core_headers;
 };
 
 struct interface {
@@ -109,6 +190,7 @@ struct arg {
 	char *interface_name;
 	struct wl_list link;
 	char *summary;
+	char *enumeration_name;
 };
 
 struct enumeration {
@@ -117,6 +199,7 @@ struct enumeration {
 	struct wl_list entry_list;
 	struct wl_list link;
 	struct description *description;
+	bool bitfield;
 };
 
 struct entry {
@@ -151,9 +234,15 @@ fail_on_null(void *p)
 }
 
 static void *
-xmalloc(size_t s)
+zalloc(size_t s)
 {
-	return fail_on_null(malloc(s));
+	return calloc(s, 1);
+}
+
+static void *
+xzalloc(size_t s)
+{
+	return fail_on_null(zalloc(s));
 }
 
 static char *
@@ -193,7 +282,7 @@ static const char *indent(int n)
 }
 
 static void
-desc_dump(char *desc, const char *fmt, ...) __attribute__((format(printf,2,3)));
+desc_dump(char *desc, const char *fmt, ...) WL_PRINTF(2, 3);
 
 static void
 desc_dump(char *desc, const char *fmt, ...)
@@ -283,7 +372,7 @@ warn(struct location *loc, const char *msg, ...)
 	va_end(ap);
 }
 
-static int
+static bool
 is_nullable_type(struct arg *arg)
 {
 	switch (arg->type) {
@@ -292,12 +381,212 @@ is_nullable_type(struct arg *arg)
 	case OBJECT:
 	case NEW_ID:
 	case ARRAY:
-		return 1;
+		return true;
 	default:
-		return 0;
+		return false;
 	}
 }
 
+static struct message *
+create_message(struct location loc, const char *name)
+{
+	struct message *message;
+
+	message = xzalloc(sizeof *message);
+	message->loc = loc;
+	message->name = xstrdup(name);
+	message->uppercase_name = uppercase_dup(name);
+	wl_list_init(&message->arg_list);
+
+	return message;
+}
+
+static void
+free_arg(struct arg *arg)
+{
+	free(arg->name);
+	free(arg->interface_name);
+	free(arg->summary);
+	free(arg);
+}
+
+static struct arg *
+create_arg(const char *name)
+{
+	struct arg *arg;
+
+	arg = xzalloc(sizeof *arg);
+	arg->name = xstrdup(name);
+
+	return arg;
+}
+
+static bool
+set_arg_type(struct arg *arg, const char *type)
+{
+	if (strcmp(type, "int") == 0)
+		arg->type = INT;
+	else if (strcmp(type, "uint") == 0)
+		arg->type = UNSIGNED;
+	else if (strcmp(type, "fixed") == 0)
+		arg->type = FIXED;
+	else if (strcmp(type, "string") == 0)
+		arg->type = STRING;
+	else if (strcmp(type, "array") == 0)
+		arg->type = ARRAY;
+	else if (strcmp(type, "fd") == 0)
+		arg->type = FD;
+	else if (strcmp(type, "new_id") == 0)
+		arg->type = NEW_ID;
+	else if (strcmp(type, "object") == 0)
+		arg->type = OBJECT;
+	else
+		return false;
+
+	return true;
+}
+
+static void
+free_description(struct description *desc)
+{
+	if (!desc)
+		return;
+
+	free(desc->summary);
+	free(desc->text);
+
+	free(desc);
+}
+
+static void
+free_message(struct message *message)
+{
+	struct arg *a, *a_next;
+
+	free(message->name);
+	free(message->uppercase_name);
+	free_description(message->description);
+
+	wl_list_for_each_safe(a, a_next, &message->arg_list, link)
+		free_arg(a);
+
+	free(message);
+}
+
+static struct enumeration *
+create_enumeration(const char *name)
+{
+	struct enumeration *enumeration;
+
+	enumeration = xzalloc(sizeof *enumeration);
+	enumeration->name = xstrdup(name);
+	enumeration->uppercase_name = uppercase_dup(name);
+
+	wl_list_init(&enumeration->entry_list);
+
+	return enumeration;
+}
+
+static struct entry *
+create_entry(const char *name, const char *value)
+{
+	struct entry *entry;
+
+	entry = xzalloc(sizeof *entry);
+	entry->name = xstrdup(name);
+	entry->uppercase_name = uppercase_dup(name);
+	entry->value = xstrdup(value);
+
+	return entry;
+}
+
+static void
+free_entry(struct entry *entry)
+{
+	free(entry->name);
+	free(entry->uppercase_name);
+	free(entry->value);
+	free(entry->summary);
+
+	free(entry);
+}
+
+static void
+free_enumeration(struct enumeration *enumeration)
+{
+	struct entry *e, *e_next;
+
+	free(enumeration->name);
+	free(enumeration->uppercase_name);
+	free_description(enumeration->description);
+
+	wl_list_for_each_safe(e, e_next, &enumeration->entry_list, link)
+		free_entry(e);
+
+	free(enumeration);
+}
+
+static struct interface *
+create_interface(struct location loc, const char *name, int version)
+{
+	struct interface *interface;
+
+	interface = xzalloc(sizeof *interface);
+	interface->loc = loc;
+	interface->name = xstrdup(name);
+	interface->uppercase_name = uppercase_dup(name);
+	interface->version = version;
+	interface->since = 1;
+	wl_list_init(&interface->request_list);
+	wl_list_init(&interface->event_list);
+	wl_list_init(&interface->enumeration_list);
+
+	return interface;
+}
+
+static void
+free_interface(struct interface *interface)
+{
+	struct message *m, *next_m;
+	struct enumeration *e, *next_e;
+
+	free(interface->name);
+	free(interface->uppercase_name);
+	free_description(interface->description);
+
+	wl_list_for_each_safe(m, next_m, &interface->request_list, link)
+		free_message(m);
+	wl_list_for_each_safe(m, next_m, &interface->event_list, link)
+		free_message(m);
+	wl_list_for_each_safe(e, next_e, &interface->enumeration_list, link)
+		free_enumeration(e);
+
+	free(interface);
+}
+
+/* convert string to unsigned integer,
+ * in the case of error, return -1 */
+static int
+strtouint(const char *str)
+{
+	long int ret;
+	char *end;
+	int prev_errno = errno;
+
+	errno = 0;
+	ret = strtol(str, &end, 10);
+	if (errno != 0 || end == str || *end != '\0')
+		return -1;
+
+	/* check range */
+	if (ret < 0 || ret > INT_MAX) {
+		return -1;
+	}
+
+	errno = prev_errno;
+	return (int)ret;
+}
+
 static void
 start_element(void *data, const char *element_name, const char **atts)
 {
@@ -307,27 +596,27 @@ start_element(void *data, const char *element_name, const char **atts)
 	struct arg *arg;
 	struct enumeration *enumeration;
 	struct entry *entry;
-	struct description *description;
-	const char *name, *type, *interface_name, *value, *summary, *since;
-	const char *allow_null;
-	char *end;
-	int i, version;
+	struct description *description = NULL;
+	const char *name = NULL;
+	const char *type = NULL;
+	const char *interface_name = NULL;
+	const char *value = NULL;
+	const char *summary = NULL;
+	const char *since = NULL;
+	const char *allow_null = NULL;
+	const char *enumeration_name = NULL;
+	const char *bitfield = NULL;
+	int i, version = 0;
 
 	ctx->loc.line_number = XML_GetCurrentLineNumber(ctx->parser);
-	name = NULL;
-	type = NULL;
-	version = 0;
-	interface_name = NULL;
-	value = NULL;
-	summary = NULL;
-	description = NULL;
-	since = NULL;
-	allow_null = NULL;
 	for (i = 0; atts[i]; i += 2) {
 		if (strcmp(atts[i], "name") == 0)
 			name = atts[i + 1];
-		if (strcmp(atts[i], "version") == 0)
-			version = atoi(atts[i + 1]);
+		if (strcmp(atts[i], "version") == 0) {
+			version = strtouint(atts[i + 1]);
+			if (version == -1)
+				fail(&ctx->loc, "wrong version (%s)", atts[i + 1]);
+		}
 		if (strcmp(atts[i], "type") == 0)
 			type = atts[i + 1];
 		if (strcmp(atts[i], "value") == 0)
@@ -340,6 +629,10 @@ start_element(void *data, const char *element_name, const char **atts)
 			since = atts[i + 1];
 		if (strcmp(atts[i], "allow-null") == 0)
 			allow_null = atts[i + 1];
+		if (strcmp(atts[i], "enum") == 0)
+			enumeration_name = atts[i + 1];
+		if (strcmp(atts[i], "bitfield") == 0)
+			bitfield = atts[i + 1];
 	}
 
 	ctx->character_data_length = 0;
@@ -349,9 +642,8 @@ start_element(void *data, const char *element_name, const char **atts)
 
 		ctx->protocol->name = xstrdup(name);
 		ctx->protocol->uppercase_name = uppercase_dup(name);
-		ctx->protocol->description = NULL;
 	} else if (strcmp(element_name, "copyright") == 0) {
-		
+
 	} else if (strcmp(element_name, "interface") == 0) {
 		if (name == NULL)
 			fail(&ctx->loc, "no interface name given");
@@ -359,32 +651,16 @@ start_element(void *data, const char *element_name, const char **atts)
 		if (version == 0)
 			fail(&ctx->loc, "no interface version given");
 
-		interface = xmalloc(sizeof *interface);
-		interface->loc = ctx->loc;
-		interface->name = xstrdup(name);
-		interface->uppercase_name = uppercase_dup(name);
-		interface->version = version;
-		interface->description = NULL;
-		interface->since = 1;
-		wl_list_init(&interface->request_list);
-		wl_list_init(&interface->event_list);
-		wl_list_init(&interface->enumeration_list);
+		interface = create_interface(ctx->loc, name, version);
+		ctx->interface = interface;
 		wl_list_insert(ctx->protocol->interface_list.prev,
 			       &interface->link);
-		ctx->interface = interface;
 	} else if (strcmp(element_name, "request") == 0 ||
 		   strcmp(element_name, "event") == 0) {
 		if (name == NULL)
 			fail(&ctx->loc, "no request name given");
 
-		message = xmalloc(sizeof *message);
-		message->loc = ctx->loc;
-		message->name = xstrdup(name);
-		message->uppercase_name = uppercase_dup(name);
-		wl_list_init(&message->arg_list);
-		message->arg_count = 0;
-		message->new_id_count = 0;
-		message->description = NULL;
+		message = create_message(ctx->loc, name);
 
 		if (strcmp(element_name, "request") == 0)
 			wl_list_insert(ctx->interface->request_list.prev,
@@ -395,21 +671,23 @@ start_element(void *data, const char *element_name, const char **atts)
 
 		if (type != NULL && strcmp(type, "destructor") == 0)
 			message->destructor = 1;
-		else
-			message->destructor = 0;
 
 		if (since != NULL) {
-			errno = 0;
-			version = strtol(since, &end, 0);
-			if (errno == EINVAL || end == since || *end != '\0')
-				fail(&ctx->loc,
-				     "invalid integer (%s)\n", since);
-			if (version < ctx->interface->since)
-				fail(&ctx->loc, "since version not increasing\n");
-			ctx->interface->since = version;
+			version = strtouint(since);
+			if (version == -1) {
+				fail(&ctx->loc, "invalid integer (%s)\n", since);
+			} else if (version > ctx->interface->version) {
+				fail(&ctx->loc, "since (%u) larger than version (%u)\n",
+				     version, ctx->interface->version);
+			}
+		} else {
+			version = 1;
 		}
 
-		message->since = ctx->interface->since;
+		if (version < ctx->interface->since)
+			warn(&ctx->loc, "since version not increasing\n");
+		ctx->interface->since = version;
+		message->since = version;
 
 		if (strcmp(name, "destroy") == 0 && !message->destructor)
 			fail(&ctx->loc, "destroy request should be destructor type");
@@ -419,28 +697,9 @@ start_element(void *data, const char *element_name, const char **atts)
 		if (name == NULL)
 			fail(&ctx->loc, "no argument name given");
 
-		arg = xmalloc(sizeof *arg);
-		arg->name = xstrdup(name);
-
-		if (strcmp(type, "int") == 0)
-			arg->type = INT;
-		else if (strcmp(type, "uint") == 0)
-			arg->type = UNSIGNED;
-		else if (strcmp(type, "fixed") == 0)
-			arg->type = FIXED;
-		else if (strcmp(type, "string") == 0)
-			arg->type = STRING;
-		else if (strcmp(type, "array") == 0)
-			arg->type = ARRAY;
-		else if (strcmp(type, "fd") == 0)
-			arg->type = FD;
-		else if (strcmp(type, "new_id") == 0) {
-			arg->type = NEW_ID;
-		} else if (strcmp(type, "object") == 0) {
-			arg->type = OBJECT;
-		} else {
+		arg = create_arg(name);
+		if (!set_arg_type(arg, type))
 			fail(&ctx->loc, "unknown type (%s)", type);
-		}
 
 		switch (arg->type) {
 		case NEW_ID:
@@ -451,8 +710,6 @@ start_element(void *data, const char *element_name, const char **atts)
 		case OBJECT:
 			if (interface_name)
 				arg->interface_name = xstrdup(interface_name);
-			else
-				arg->interface_name = NULL;
 			break;
 		default:
 			if (interface_name != NULL)
@@ -460,17 +717,24 @@ start_element(void *data, const char *element_name, const char **atts)
 			break;
 		}
 
-		if (allow_null == NULL || strcmp(allow_null, "false") == 0)
-			arg->nullable = 0;
-		else if (strcmp(allow_null, "true") == 0)
-			arg->nullable = 1;
+		if (allow_null) {
+			if (strcmp(allow_null, "true") == 0)
+				arg->nullable = 1;
+			else if (strcmp(allow_null, "false") != 0)
+				fail(&ctx->loc,
+				     "invalid value for allow-null attribute (%s)",
+				     allow_null);
+
+			if (!is_nullable_type(arg))
+				fail(&ctx->loc,
+				     "allow-null is only valid for objects, strings, and arrays");
+		}
+
+		if (enumeration_name == NULL || strcmp(enumeration_name, "") == 0)
+			arg->enumeration_name = NULL;
 		else
-			fail(&ctx->loc, "invalid value for allow-null attribute (%s)", allow_null);
+			arg->enumeration_name = xstrdup(enumeration_name);
 
-		if (allow_null != NULL && !is_nullable_type(arg))
-			fail(&ctx->loc, "allow-null is only valid for objects, strings, and arrays");
-
-		arg->summary = NULL;
 		if (summary)
 			arg->summary = xstrdup(summary);
 
@@ -480,11 +744,16 @@ start_element(void *data, const char *element_name, const char **atts)
 		if (name == NULL)
 			fail(&ctx->loc, "no enum name given");
 
-		enumeration = xmalloc(sizeof *enumeration);
-		enumeration->name = xstrdup(name);
-		enumeration->uppercase_name = uppercase_dup(name);
-		enumeration->description = NULL;
-		wl_list_init(&enumeration->entry_list);
+		enumeration = create_enumeration(name);
+
+		if (bitfield == NULL || strcmp(bitfield, "false") == 0)
+			enumeration->bitfield = false;
+		else if (strcmp(bitfield, "true") == 0)
+			enumeration->bitfield = true;
+		else
+			fail(&ctx->loc,
+                             "invalid value (%s) for bitfield attribute (only true/false are accepted)",
+                             bitfield);
 
 		wl_list_insert(ctx->interface->enumeration_list.prev,
 			       &enumeration->link);
@@ -494,10 +763,8 @@ start_element(void *data, const char *element_name, const char **atts)
 		if (name == NULL)
 			fail(&ctx->loc, "no entry name given");
 
-		entry = xmalloc(sizeof *entry);
-		entry->name = xstrdup(name);
-		entry->uppercase_name = uppercase_dup(name);
-		entry->value = xstrdup(value);
+		entry = create_entry(name, value);
+
 		if (summary)
 			entry->summary = xstrdup(summary);
 		else
@@ -508,7 +775,7 @@ start_element(void *data, const char *element_name, const char **atts)
 		if (summary == NULL)
 			fail(&ctx->loc, "description without summary");
 
-		description = xmalloc(sizeof *description);
+		description = xzalloc(sizeof *description);
 		description->summary = xstrdup(summary);
 
 		if (ctx->message)
@@ -523,6 +790,46 @@ start_element(void *data, const char *element_name, const char **atts)
 	}
 }
 
+static void
+verify_arguments(struct parse_context *ctx, struct wl_list *messages, struct wl_list *enumerations)
+{
+	struct message *m;
+	wl_list_for_each(m, messages, link) {
+		struct arg *a;
+		wl_list_for_each(a, &m->arg_list, link) {
+			struct enumeration *e, *f;
+
+			if (!a->enumeration_name)
+				continue;
+
+			f = NULL;
+			wl_list_for_each(e, enumerations, link) {
+				if(strcmp(e->name, a->enumeration_name) == 0)
+					f = e;
+			}
+
+			if (f == NULL)
+				fail(&ctx->loc,
+				     "could not find enumeration %s",
+				     a->enumeration_name);
+
+			switch (a->type) {
+			case INT:
+				if (f->bitfield)
+					fail(&ctx->loc,
+					     "bitfield-style enum must only be referenced by uint");
+				break;
+			case UNSIGNED:
+				break;
+			default:
+				fail(&ctx->loc,
+				     "enumeration-style argument has wrong type");
+			}
+		}
+	}
+
+}
+
 static void
 end_element(void *data, const XML_Char *name)
 {
@@ -541,7 +848,17 @@ end_element(void *data, const XML_Char *name)
 		   strcmp(name, "event") == 0) {
 		ctx->message = NULL;
 	} else if (strcmp(name, "enum") == 0) {
+		if (wl_list_empty(&ctx->enumeration->entry_list)) {
+			fail(&ctx->loc, "enumeration %s was empty",
+			     ctx->enumeration->name);
+		}
 		ctx->enumeration = NULL;
+	} else if (strcmp(name, "interface") == 0) {
+		struct interface *i = ctx->interface;
+
+		verify_arguments(ctx, &i->request_list, &i->enumeration_list);
+		verify_arguments(ctx, &i->event_list, &i->enumeration_list);
+
 	}
 }
 
@@ -576,6 +893,18 @@ emit_opcodes(struct wl_list *message_list, struct interface *interface)
 	printf("\n");
 }
 
+static void
+emit_opcode_versions(struct wl_list *message_list, struct interface *interface)
+{
+	struct message *m;
+
+	wl_list_for_each(m, message_list, link)
+		printf("#define %s_%s_SINCE_VERSION\t%d\n",
+		       interface->uppercase_name, m->uppercase_name, m->since);
+
+	printf("\n");
+}
+
 static void
 emit_type(struct arg *a)
 {
@@ -627,6 +956,14 @@ emit_stubs(struct wl_list *message_list, struct interface *interface)
 	       interface->name, interface->name, interface->name,
 	       interface->name);
 
+	printf("static inline uint32_t\n"
+	       "%s_get_version(struct %s *%s)\n"
+	       "{\n"
+	       "\treturn wl_proxy_get_version((struct wl_proxy *) %s);\n"
+	       "}\n\n",
+	       interface->name, interface->name, interface->name,
+	       interface->name);
+
 	has_destructor = 0;
 	has_destroy = 0;
 	wl_list_for_each(m, message_list, link) {
@@ -698,21 +1035,31 @@ emit_stubs(struct wl_list *message_list, struct interface *interface)
 
 		printf(")\n"
 		       "{\n");
-		if (ret) {
+		if (ret && ret->interface_name == NULL) {
+			/* an arg has type ="new_id" but interface is not
+			 * provided, such as in wl_registry.bind */
 			printf("\tstruct wl_proxy *%s;\n\n"
-			       "\t%s = wl_proxy_marshal_constructor("
+			       "\t%s = wl_proxy_marshal_constructor_versioned("
 			       "(struct wl_proxy *) %s,\n"
-			       "\t\t\t %s_%s, ",
+			       "\t\t\t %s_%s, interface, version",
 			       ret->name, ret->name,
 			       interface->name,
 			       interface->uppercase_name,
 			       m->uppercase_name);
-
-			if (ret->interface_name == NULL)
-				printf("interface");
-			else
-				printf("&%s_interface", ret->interface_name);
+		} else if (ret) {
+			/* Normal factory case, an arg has type="new_id" and
+			 * an interface is provided */
+			printf("\tstruct wl_proxy *%s;\n\n"
+			       "\t%s = wl_proxy_marshal_constructor("
+			       "(struct wl_proxy *) %s,\n"
+			       "\t\t\t %s_%s, &%s_interface",
+			       ret->name, ret->name,
+			       interface->name,
+			       interface->uppercase_name,
+			       m->uppercase_name,
+			       ret->interface_name);
 		} else {
+			/* No args have type="new_id" */
 			printf("\twl_proxy_marshal((struct wl_proxy *) %s,\n"
 			       "\t\t\t %s_%s",
 			       interface->name,
@@ -832,16 +1179,15 @@ emit_enumerations(struct interface *interface)
 }
 
 static void
-emit_structs(struct wl_list *message_list, struct interface *interface)
+emit_structs(struct wl_list *message_list, struct interface *interface, enum side side)
 {
 	struct message *m;
 	struct arg *a;
-	int is_interface, n;
+	int n;
 
 	if (wl_list_empty(message_list))
 		return;
 
-	is_interface = message_list == &interface->request_list;
 	if (interface->description) {
 		struct description *desc = interface->description;
 		printf("/**\n");
@@ -857,7 +1203,7 @@ emit_structs(struct wl_list *message_list, struct interface *interface)
 		printf(" */\n");
 	}
 	printf("struct %s_%s {\n", interface->name,
-	       is_interface ? "interface" : "listener");
+	       (side == SERVER) ? "interface" : "listener");
 
 	wl_list_for_each(m, message_list, link) {
 		struct description *mdesc = m->description;
@@ -866,8 +1212,7 @@ emit_structs(struct wl_list *message_list, struct interface *interface)
 		desc_dump(mdesc ? mdesc->summary : "(none)",
 			  "\t * %s - ", m->name);
 		wl_list_for_each(a, &m->arg_list, link) {
-
-			if (is_interface && a->type == NEW_ID &&
+			if (side == SERVER && a->type == NEW_ID &&
 			    a->interface_name == NULL)
 				printf("\t * @interface: name of the objects interface\n"
 				       "\t * @version: version of the objects interface\n");
@@ -887,7 +1232,7 @@ emit_structs(struct wl_list *message_list, struct interface *interface)
 		printf("\tvoid (*%s)(", m->name);
 
 		n = strlen(m->name) + 17;
-		if (is_interface) {
+		if (side == SERVER) {
 			printf("struct wl_client *client,\n"
 			       "%sstruct wl_resource *resource",
 			       indent(n));
@@ -900,14 +1245,14 @@ emit_structs(struct wl_list *message_list, struct interface *interface)
 		wl_list_for_each(a, &m->arg_list, link) {
 			printf(",\n%s", indent(n));
 
-			if (is_interface && a->type == OBJECT)
+			if (side == SERVER && a->type == OBJECT)
 				printf("struct wl_resource *");
-			else if (is_interface && a->type == NEW_ID && a->interface_name == NULL)
+			else if (side == SERVER && a->type == NEW_ID && a->interface_name == NULL)
 				printf("const char *interface, uint32_t version, uint32_t ");
-			else if (!is_interface && a->type == OBJECT && a->interface_name == NULL)
+			else if (side == CLIENT && a->type == OBJECT && a->interface_name == NULL)
 				printf("void *");
 
-			else if (!is_interface && a->type == NEW_ID)
+			else if (side == CLIENT && a->type == NEW_ID)
 				printf("struct %s *", a->interface_name);
 			else
 				emit_type(a);
@@ -920,7 +1265,7 @@ emit_structs(struct wl_list *message_list, struct interface *interface)
 
 	printf("};\n\n");
 
-	if (!is_interface) {
+	if (side == CLIENT) {
 	    printf("static inline int\n"
 		   "%s_add_listener(struct %s *%s,\n"
 		   "%sconst struct %s_listener *listener, void *data)\n"
@@ -939,31 +1284,90 @@ emit_structs(struct wl_list *message_list, struct interface *interface)
 static void
 format_copyright(const char *copyright)
 {
-	int bol = 1, start = 0, i;
+	int bol = 1, start = 0, i, length;
+	bool comment_started = false;
 
-	for (i = 0; copyright[i]; i++) {
+	length = strlen(copyright);
+	for (i = 0; i <= length; i++) {
 		if (bol && (copyright[i] == ' ' || copyright[i] == '\t')) {
 			continue;
 		} else if (bol) {
 			bol = 0;
 			start = i;
 		}
-
-		if (copyright[i] == '\n' || copyright[i] == '\0') {
-			printf("%s %.*s\n",
-			       i == 0 ? "/*" : " *",
+		if (copyright[i] == '\n' ||
+		    (copyright[i] == '\0' && !(start == i))) {
+			printf("%s%s%.*s\n",
+			       comment_started ? " *" : "/*",
+			       i > start ? " " : "",
 			       i - start, copyright + start);
 			bol = 1;
+			comment_started = true;
 		}
 	}
-	printf(" */\n\n");
+	if (comment_started)
+		printf(" */\n\n");
 }
 
 static void
-emit_header(struct protocol *protocol, int server)
+emit_types_forward_declarations(struct protocol *protocol,
+				struct wl_list *message_list,
+				struct wl_array *types)
 {
-	struct interface *i;
-	const char *s = server ? "SERVER" : "CLIENT";
+	struct message *m;
+	struct arg *a;
+	int length;
+	char **p;
+
+	wl_list_for_each(m, message_list, link) {
+		length = 0;
+		m->all_null = 1;
+		wl_list_for_each(a, &m->arg_list, link) {
+			length++;
+			switch (a->type) {
+			case NEW_ID:
+			case OBJECT:
+				if (!a->interface_name)
+					continue;
+
+				m->all_null = 0;
+				p = fail_on_null(wl_array_add(types, sizeof *p));
+				*p = a->interface_name;
+				break;
+			default:
+				break;
+			}
+		}
+
+		if (m->all_null && length > protocol->null_run_length)
+			protocol->null_run_length = length;
+	}
+}
+
+static int
+cmp_names(const void *p1, const void *p2)
+{
+	const char * const *s1 = p1, * const *s2 = p2;
+
+	return strcmp(*s1, *s2);
+}
+
+static const char *
+get_include_name(bool core, enum side side)
+{
+	if (side == SERVER)
+		return core ? "wayland-server-core.h" : "wayland-server.h";
+	else
+		return core ? "wayland-client-core.h" : "wayland-client.h";
+}
+
+static void
+emit_header(struct protocol *protocol, enum side side)
+{
+	struct interface *i, *i_next;
+	struct wl_array types;
+	const char *s = (side == SERVER) ? "SERVER" : "CLIENT";
+	char **p, *prev;
 
 	if (protocol->copyright)
 		format_copyright(protocol->copyright);
@@ -982,32 +1386,54 @@ emit_header(struct protocol *protocol, int server)
 	       "struct wl_resource;\n\n",
 	       protocol->uppercase_name, s,
 	       protocol->uppercase_name, s,
-	       server ? "wayland-util.h" : "wayland-client.h");
+	       get_include_name(protocol->core_headers, side));
 
-	wl_list_for_each(i, &protocol->interface_list, link)
-		printf("struct %s;\n", i->name);
+	wl_array_init(&types);
+	wl_list_for_each(i, &protocol->interface_list, link) {
+		emit_types_forward_declarations(protocol, &i->request_list, &types);
+		emit_types_forward_declarations(protocol, &i->event_list, &types);
+	}
+
+	wl_list_for_each(i, &protocol->interface_list, link) {
+		p = fail_on_null(wl_array_add(&types, sizeof *p));
+		*p = i->name;
+	}
+
+	qsort(types.data, types.size / sizeof *p, sizeof *p, cmp_names);
+	prev = NULL;
+	wl_array_for_each(p, &types) {
+		if (prev && strcmp(*p, prev) == 0)
+			continue;
+		printf("struct %s;\n", *p);
+		prev = *p;
+	}
+	wl_array_release(&types);
 	printf("\n");
 
 	wl_list_for_each(i, &protocol->interface_list, link) {
 		printf("extern const struct wl_interface "
-		       "%s_interface;\n",
-		       i->name);
+		       "%s_interface;\n", i->name);
 	}
+
 	printf("\n");
 
-	wl_list_for_each(i, &protocol->interface_list, link) {
+	wl_list_for_each_safe(i, i_next, &protocol->interface_list, link) {
 
 		emit_enumerations(i);
 
-		if (server) {
-			emit_structs(&i->request_list, i);
+		if (side == SERVER) {
+			emit_structs(&i->request_list, i, side);
 			emit_opcodes(&i->event_list, i);
+			emit_opcode_versions(&i->event_list, i);
 			emit_event_wrappers(&i->event_list, i);
 		} else {
-			emit_structs(&i->event_list, i);
+			emit_structs(&i->event_list, i, side);
 			emit_opcodes(&i->request_list, i);
+			emit_opcode_versions(&i->request_list, i);
 			emit_stubs(&i->request_list, i);
 		}
+
+		free_interface(i);
 	}
 
 	printf("#ifdef  __cplusplus\n"
@@ -1017,39 +1443,6 @@ emit_header(struct protocol *protocol, int server)
 	       "#endif\n");
 }
 
-static void
-emit_types_forward_declarations(struct protocol *protocol,
-				struct wl_list *message_list)
-{
-	struct message *m;
-	struct arg *a;
-	int length;
-
-	wl_list_for_each(m, message_list, link) {
-		length = 0;
-		m->all_null = 1;
-		wl_list_for_each(a, &m->arg_list, link) {
-			length++;
-			switch (a->type) {
-			case NEW_ID:
-			case OBJECT:
-				if (!a->interface_name)
-					continue;
-
-				m->all_null = 0;
-				printf("extern const struct wl_interface %s_interface;\n",
-				       a->interface_name);
-				break;
-			default:
-				break;
-			}
-		}
-
-		if (m->all_null && length > protocol->null_run_length)
-			protocol->null_run_length = length;
-	}
-}
-
 static void
 emit_null_run(struct protocol *protocol)
 {
@@ -1156,7 +1549,9 @@ emit_messages(struct wl_list *message_list,
 static void
 emit_code(struct protocol *protocol)
 {
-	struct interface *i;
+	struct interface *i, *next;
+	struct wl_array types;
+	char **p, *prev;
 
 	if (protocol->copyright)
 		format_copyright(protocol->copyright);
@@ -1165,10 +1560,20 @@ emit_code(struct protocol *protocol)
 	       "#include <stdint.h>\n"
 	       "#include \"wayland-util.h\"\n\n");
 
+	wl_array_init(&types);
 	wl_list_for_each(i, &protocol->interface_list, link) {
-		emit_types_forward_declarations(protocol, &i->request_list);
-		emit_types_forward_declarations(protocol, &i->event_list);
+		emit_types_forward_declarations(protocol, &i->request_list, &types);
+		emit_types_forward_declarations(protocol, &i->event_list, &types);
 	}
+	qsort(types.data, types.size / sizeof *p, sizeof *p, cmp_names);
+	prev = NULL;
+	wl_array_for_each(p, &types) {
+		if (prev && strcmp(*p, prev) == 0)
+			continue;
+		printf("extern const struct wl_interface %s_interface;\n", *p);
+		prev = *p;
+	}
+	wl_array_release(&types);
 	printf("\n");
 
 	printf("static const struct wl_interface *types[] = {\n");
@@ -1179,7 +1584,7 @@ emit_code(struct protocol *protocol)
 	}
 	printf("};\n\n");
 
-	wl_list_for_each(i, &protocol->interface_list, link) {
+	wl_list_for_each_safe(i, next, &protocol->interface_list, link) {
 
 		emit_messages(&i->request_list, i, "requests");
 		emit_messages(&i->event_list, i, "events");
@@ -1202,46 +1607,125 @@ emit_code(struct protocol *protocol)
 			printf("\t0, NULL,\n");
 
 		printf("};\n\n");
+
+		/* we won't need it any further */
+		free_interface(i);
 	}
 }
 
+static void
+free_protocol(struct protocol *protocol)
+{
+	free(protocol->name);
+	free(protocol->uppercase_name);
+	free(protocol->copyright);
+	free_description(protocol->description);
+}
+
 int main(int argc, char *argv[])
 {
 	struct parse_context ctx;
 	struct protocol protocol;
+	FILE *input = stdin;
+	char *input_filename = NULL;
 	int len;
 	void *buf;
+	bool help = false, core_headers = false;
+	bool fail = false;
+	int opt;
 	enum {
 		CLIENT_HEADER,
 		SERVER_HEADER,
 		CODE,
 	} mode;
 
-	if (argc != 2)
-		usage(EXIT_FAILURE);
-	else if (strcmp(argv[1], "help") == 0 || strcmp(argv[1], "--help") == 0)
+	static const struct option options[] = {
+		{ "help",              no_argument, NULL, 'h' },
+		{ "include-core-only", no_argument, NULL, 'c' },
+		{ 0,                   0,           NULL, 0 }
+	};
+
+	while (1) {
+		opt = getopt_long(argc, argv, "hc", options, NULL);
+
+		if (opt == -1)
+			break;
+
+		switch (opt) {
+		case 'h':
+			help = true;
+			break;
+		case 'c':
+			core_headers = true;
+			break;
+		default:
+			fail = true;
+			break;
+		}
+	}
+
+	argv += optind;
+	argc -= optind;
+
+	if (help)
 		usage(EXIT_SUCCESS);
-	else if (strcmp(argv[1], "client-header") == 0)
+	else if ((argc != 1 && argc != 3) || fail)
+		usage(EXIT_FAILURE);
+	else if (strcmp(argv[0], "help") == 0)
+		usage(EXIT_SUCCESS);
+	else if (strcmp(argv[0], "client-header") == 0)
 		mode = CLIENT_HEADER;
-	else if (strcmp(argv[1], "server-header") == 0)
+	else if (strcmp(argv[0], "server-header") == 0)
 		mode = SERVER_HEADER;
-	else if (strcmp(argv[1], "code") == 0)
+	else if (strcmp(argv[0], "code") == 0)
 		mode = CODE;
 	else
 		usage(EXIT_FAILURE);
 
+	if (argc == 3) {
+		input_filename = argv[1];
+		input = fopen(input_filename, "r");
+		if (input == NULL) {
+			fprintf(stderr, "Could not open input file: %s\n",
+				strerror(errno));
+			exit(EXIT_FAILURE);
+		}
+		if (freopen(argv[2], "w", stdout) == NULL) {
+			fprintf(stderr, "Could not open output file: %s\n",
+				strerror(errno));
+			fclose(input);
+			exit(EXIT_FAILURE);
+		}
+	}
+
+	/* initialize protocol structure */
+	memset(&protocol, 0, sizeof protocol);
 	wl_list_init(&protocol.interface_list);
-	protocol.type_index = 0;
-	protocol.null_run_length = 0;
-	protocol.copyright = NULL;
+	protocol.core_headers = core_headers;
+
+	/* initialize context */
 	memset(&ctx, 0, sizeof ctx);
 	ctx.protocol = &protocol;
+	if (input == stdin)
+		ctx.loc.filename = "<stdin>";
+	else
+		ctx.loc.filename = input_filename;
 
-	ctx.loc.filename = "<stdin>";
+	if (!is_dtd_valid(input, ctx.loc.filename)) {
+		fprintf(stderr,
+		"*******************************************************\n"
+		"*                                                     *\n"
+		"* WARNING: XML failed validation against built-in DTD *\n"
+		"*                                                     *\n"
+		"*******************************************************\n");
+	}
+
+	/* create XML parser */
 	ctx.parser = XML_ParserCreate(NULL);
 	XML_SetUserData(ctx.parser, &ctx);
 	if (ctx.parser == NULL) {
 		fprintf(stderr, "failed to create parser\n");
+		fclose(input);
 		exit(EXIT_FAILURE);
 	}
 
@@ -1250,28 +1734,39 @@ int main(int argc, char *argv[])
 
 	do {
 		buf = XML_GetBuffer(ctx.parser, XML_BUFFER_SIZE);
-		len = fread(buf, 1, XML_BUFFER_SIZE, stdin);
+		len = fread(buf, 1, XML_BUFFER_SIZE, input);
 		if (len < 0) {
 			fprintf(stderr, "fread: %m\n");
+			fclose(input);
+			exit(EXIT_FAILURE);
+		}
+		if (XML_ParseBuffer(ctx.parser, len, len == 0) == 0) {
+			fprintf(stderr,
+				"Error parsing XML at line %ld col %ld: %s\n",
+				XML_GetCurrentLineNumber(ctx.parser),
+				XML_GetCurrentColumnNumber(ctx.parser),
+				XML_ErrorString(XML_GetErrorCode(ctx.parser)));
+			fclose(input);
 			exit(EXIT_FAILURE);
 		}
-		XML_ParseBuffer(ctx.parser, len, len == 0);
-
 	} while (len > 0);
 
 	XML_ParserFree(ctx.parser);
 
 	switch (mode) {
 		case CLIENT_HEADER:
-			emit_header(&protocol, 0);
+			emit_header(&protocol, CLIENT);
 			break;
 		case SERVER_HEADER:
-			emit_header(&protocol, 1);
+			emit_header(&protocol, SERVER);
 			break;
 		case CODE:
 			emit_code(&protocol);
 			break;
 	}
 
+	free_protocol(&protocol);
+	fclose(input);
+
 	return 0;
 }
