@@ -3,29 +3,33 @@
  * Copyright © 2011 Intel Corporation
  * Copyright © 2013 Jason Ekstrand
  *
- * Permission to use, copy, modify, distribute, and sell this software and its
- * documentation for any purpose is hereby granted without fee, provided that
- * the above copyright notice appear in all copies and that both that copyright
- * notice and this permission notice appear in supporting documentation, and
- * that the name of the copyright holders not be used in advertising or
- * publicity pertaining to distribution of the software without specific,
- * written prior permission.  The copyright holders make no representations
- * about the suitability of this software for any purpose.  It is provided "as
- * is" without express or implied warranty.
+ * Permission is hereby granted, free of charge, to any person obtaining
+ * a copy of this software and associated documentation files (the
+ * "Software"), to deal in the Software without restriction, including
+ * without limitation the rights to use, copy, modify, merge, publish,
+ * distribute, sublicense, and/or sell copies of the Software, and to
+ * permit persons to whom the Software is furnished to do so, subject to
+ * the following conditions:
  *
- * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
- * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
- * EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL, INDIRECT OR
- * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
- * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
- * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
- * OF THIS SOFTWARE.
+ * The above copyright notice and this permission notice (including the
+ * next paragraph) shall be included in all copies or substantial
+ * portions of the Software.
+ *
+ * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
+ * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
+ * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
+ * NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
+ * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
+ * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
+ * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
+ * SOFTWARE.
  */
 
 #ifndef WAYLAND_PRIVATE_H
 #define WAYLAND_PRIVATE_H
 
 #include <stdarg.h>
+#include <stdlib.h>
 
 #define WL_HIDE_DEPRECATED 1
 
@@ -68,34 +72,72 @@ struct wl_map {
 
 typedef void (*wl_iterator_func_t)(void *element, void *data);
 
-void wl_map_init(struct wl_map *map, uint32_t side);
-void wl_map_release(struct wl_map *map);
-uint32_t wl_map_insert_new(struct wl_map *map, uint32_t flags, void *data);
-int wl_map_insert_at(struct wl_map *map, uint32_t flags, uint32_t i, void *data);
-int wl_map_reserve_new(struct wl_map *map, uint32_t i);
-void wl_map_remove(struct wl_map *map, uint32_t i);
-void *wl_map_lookup(struct wl_map *map, uint32_t i);
-uint32_t wl_map_lookup_flags(struct wl_map *map, uint32_t i);
-void wl_map_for_each(struct wl_map *map, wl_iterator_func_t func, void *data);
+void
+wl_map_init(struct wl_map *map, uint32_t side);
+
+void
+wl_map_release(struct wl_map *map);
+
+uint32_t
+wl_map_insert_new(struct wl_map *map, uint32_t flags, void *data);
+
+int
+wl_map_insert_at(struct wl_map *map, uint32_t flags, uint32_t i, void *data);
+
+int
+wl_map_reserve_new(struct wl_map *map, uint32_t i);
+
+void
+wl_map_remove(struct wl_map *map, uint32_t i);
+
+void *
+wl_map_lookup(struct wl_map *map, uint32_t i);
+
+uint32_t
+wl_map_lookup_flags(struct wl_map *map, uint32_t i);
+
+void
+wl_map_for_each(struct wl_map *map, wl_iterator_func_t func, void *data);
 
 struct wl_connection;
 struct wl_closure;
 struct wl_proxy;
 
-int wl_interface_equal(const struct wl_interface *iface1,
-		       const struct wl_interface *iface2);
+int
+wl_interface_equal(const struct wl_interface *iface1,
+		   const struct wl_interface *iface2);
 
-struct wl_connection *wl_connection_create(int fd);
-void wl_connection_destroy(struct wl_connection *connection);
-void wl_connection_copy(struct wl_connection *connection, void *data, size_t size);
-void wl_connection_consume(struct wl_connection *connection, size_t size);
+struct wl_connection *
+wl_connection_create(int fd);
 
-int wl_connection_flush(struct wl_connection *connection);
-int wl_connection_read(struct wl_connection *connection);
+int
+wl_connection_destroy(struct wl_connection *connection);
 
-int wl_connection_write(struct wl_connection *connection, const void *data, size_t count);
-int wl_connection_queue(struct wl_connection *connection,
-			const void *data, size_t count);
+void
+wl_connection_copy(struct wl_connection *connection, void *data, size_t size);
+
+void
+wl_connection_consume(struct wl_connection *connection, size_t size);
+
+int
+wl_connection_flush(struct wl_connection *connection);
+
+uint32_t
+wl_connection_pending_input(struct wl_connection *connection);
+
+int
+wl_connection_read(struct wl_connection *connection);
+
+int
+wl_connection_write(struct wl_connection *connection,
+		    const void *data, size_t count);
+
+int
+wl_connection_queue(struct wl_connection *connection,
+		    const void *data, size_t count);
+
+int
+wl_connection_get_fd(struct wl_connection *connection);
 
 struct wl_closure {
 	int count;
@@ -130,6 +172,7 @@ struct wl_closure *
 wl_closure_marshal(struct wl_object *sender,
 		    uint32_t opcode, union wl_argument *args,
 		    const struct wl_message *message);
+
 struct wl_closure *
 wl_closure_vmarshal(struct wl_object *sender,
 		    uint32_t opcode, va_list ap,
@@ -152,25 +195,38 @@ enum wl_closure_invoke_flag {
 void
 wl_closure_invoke(struct wl_closure *closure, uint32_t flags,
 		  struct wl_object *target, uint32_t opcode, void *data);
+
 void
 wl_closure_dispatch(struct wl_closure *closure, wl_dispatcher_func_t dispatcher,
 		    struct wl_object *target, uint32_t opcode);
+
 int
 wl_closure_send(struct wl_closure *closure, struct wl_connection *connection);
+
 int
 wl_closure_queue(struct wl_closure *closure, struct wl_connection *connection);
+
 void
-wl_closure_print(struct wl_closure *closure, struct wl_object *target, int send);
+wl_closure_print(struct wl_closure *closure,
+		 struct wl_object *target, int send);
+
 void
 wl_closure_destroy(struct wl_closure *closure);
 
 extern wl_log_func_t wl_log_handler;
 
 void wl_log(const char *fmt, ...);
+void wl_abort(const char *fmt, ...);
 
 struct wl_display;
 
 struct wl_array *
 wl_display_get_additional_shm_formats(struct wl_display *display);
 
+static inline void *
+zalloc(size_t s)
+{
+	return calloc(1, s);
+}
+
 #endif
