@@ -1,23 +1,26 @@
 /*
  * Copyright © 2008 Kristian Høgsberg
  *
- * Permission to use, copy, modify, distribute, and sell this software and its
- * documentation for any purpose is hereby granted without fee, provided that
- * the above copyright notice appear in all copies and that both that copyright
- * notice and this permission notice appear in supporting documentation, and
- * that the name of the copyright holders not be used in advertising or
- * publicity pertaining to distribution of the software without specific,
- * written prior permission.  The copyright holders make no representations
- * about the suitability of this software for any purpose.  It is provided "as
- * is" without express or implied warranty.
+ * Permission is hereby granted, free of charge, to any person obtaining
+ * a copy of this software and associated documentation files (the
+ * "Software"), to deal in the Software without restriction, including
+ * without limitation the rights to use, copy, modify, merge, publish,
+ * distribute, sublicense, and/or sell copies of the Software, and to
+ * permit persons to whom the Software is furnished to do so, subject to
+ * the following conditions:
  *
- * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
- * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
- * EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL, INDIRECT OR
- * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
- * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
- * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
- * OF THIS SOFTWARE.
+ * The above copyright notice and this permission notice (including the
+ * next paragraph) shall be included in all copies or substantial
+ * portions of the Software.
+ *
+ * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
+ * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
+ * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
+ * NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
+ * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
+ * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
+ * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
+ * SOFTWARE.
  */
 
 #define _GNU_SOURCE
@@ -39,7 +42,6 @@
 #include <fcntl.h>
 #include <sys/file.h>
 #include <sys/stat.h>
-#include <ffi.h>
 
 #include "wayland-private.h"
 #include "wayland-server.h"
@@ -47,8 +49,9 @@
 #include "wayland-os.h"
 
 /* This is the size of the char array in struct sock_addr_un.
-   No Wayland socket can be created with a path longer than this,
-   including the null terminator. */
+ * No Wayland socket can be created with a path longer than this,
+ * including the null terminator.
+ */
 #ifndef UNIX_PATH_MAX
 #define UNIX_PATH_MAX	108
 #endif
@@ -63,6 +66,7 @@ struct wl_socket {
 	char lock_addr[UNIX_PATH_MAX + LOCK_SUFFIXLEN];
 	struct wl_list link;
 	struct wl_event_source *source;
+	char *display_name;
 };
 
 struct wl_client {
@@ -237,7 +241,7 @@ wl_client_connection_data(int fd, uint32_t mask, void *data)
 	const struct wl_message *message;
 	uint32_t p[2];
 	uint32_t resource_flags;
-	int opcode, size;
+	int opcode, size, since;
 	int len;
 
 	if (mask & (WL_EVENT_ERROR | WL_EVENT_HANGUP)) {
@@ -259,7 +263,7 @@ wl_client_connection_data(int fd, uint32_t mask, void *data)
 	len = 0;
 	if (mask & WL_EVENT_READABLE) {
 		len = wl_connection_read(connection);
-		if (len < 0 && errno != EAGAIN) {
+		if (len == 0 || (len < 0 && errno != EAGAIN)) {
 			wl_client_destroy(client);
 			return 1;
 		}
@@ -293,13 +297,14 @@ wl_client_connection_data(int fd, uint32_t mask, void *data)
 		}
 
 		message = &object->interface->methods[opcode];
+		since = wl_message_get_since(message);
 		if (!(resource_flags & WL_MAP_ENTRY_LEGACY) &&
-		    resource->version > 0 &&
-		    resource->version < wl_message_get_since(message)) {
+		    resource->version > 0 && resource->version < since) {
 			wl_resource_post_error(client->display_resource,
 					       WL_DISPLAY_ERROR_INVALID_METHOD,
-					       "invalid method %d, object %s@%u",
-					       opcode,
+					       "invalid method %d (since %d < %d)"
+					       ", object %s@%u",
+					       opcode, resource->version, since,
 					       object->interface->name,
 					       object->id);
 			break;
@@ -308,7 +313,6 @@ wl_client_connection_data(int fd, uint32_t mask, void *data)
 
 		closure = wl_connection_demarshal(client->connection, size,
 						  &client->objects, message);
-		len -= size;
 
 		if (closure == NULL && errno == ENOMEM) {
 			wl_resource_post_no_memory(resource);
@@ -341,6 +345,8 @@ wl_client_connection_data(int fd, uint32_t mask, void *data)
 
 		if (client->error)
 			break;
+
+		len = wl_connection_pending_input(connection);
 	}
 
 	if (client->error)
@@ -357,7 +363,7 @@ wl_client_connection_data(int fd, uint32_t mask, void *data)
  * socket later - typically when the compositor has handled all
  * requests and goes back to block in the event loop.  This function
  * flushes all queued up events for a client immediately.
- * 
+ *
  * \memberof wl_client
  */
 WL_EXPORT void
@@ -370,7 +376,7 @@ wl_client_flush(struct wl_client *client)
  *
  * \param client The client object
  * \return The display object the client is associated with.
- * 
+ *
  * \memberof wl_client
  */
 WL_EXPORT struct wl_display *
@@ -379,9 +385,8 @@ wl_client_get_display(struct wl_client *client)
 	return client->display;
 }
 
-static void
-bind_display(struct wl_client *client,
-	     void *data, uint32_t version, uint32_t id);
+static int
+bind_display(struct wl_client *client, struct wl_display *display);
 
 /** Create a client for the given file descriptor
  *
@@ -402,7 +407,7 @@ bind_display(struct wl_client *client,
  * WAYLAND_SOCKET environment variable on the client side.
  *
  * On failure this function sets errno accordingly and returns NULL.
- * 
+ *
  * \memberof wl_display
  */
 WL_EXPORT struct wl_client *
@@ -411,11 +416,10 @@ wl_client_create(struct wl_display *display, int fd)
 	struct wl_client *client;
 	socklen_t len;
 
-	client = malloc(sizeof *client);
+	client = zalloc(sizeof *client);
 	if (client == NULL)
 		return NULL;
 
-	memset(client, 0, sizeof *client);
 	client->display = display;
 	client->source = wl_event_loop_add_fd(display->loop, fd,
 					      WL_EVENT_READABLE,
@@ -439,9 +443,7 @@ wl_client_create(struct wl_display *display, int fd)
 		goto err_map;
 
 	wl_signal_init(&client->destroy_signal);
-	bind_display(client, display, 1, 1);
-
-	if (!client->display_resource)
+	if (bind_display(client, display) < 0)
 		goto err_map;
 
 	wl_list_insert(display->client_list.prev, &client->link);
@@ -474,7 +476,7 @@ err_client:
  * then connects using socketpair(), this function will return the
  * credentials for the compositor.  The credentials for the socketpair
  * are set at creation time in the compositor.
- * 
+ *
  * \memberof wl_client
  */
 WL_EXPORT void
@@ -489,6 +491,41 @@ wl_client_get_credentials(struct wl_client *client,
 		*gid = client->ucred.gid;
 }
 
+/** Get the file descriptor for the client
+ *
+ * \param client The display object
+ * \return The file descriptor to use for the connection
+ *
+ * This function returns the file descriptor for the given client.
+ *
+ * Be sure to use the file descriptor from the client for inspection only.
+ * If the caller does anything to the file descriptor that changes its state,
+ * it will likely cause problems.
+ *
+ * See also wl_client_get_credentials().
+ * It is recommended that you evaluate whether wl_client_get_credentials()
+ * can be applied to your use case instead of this function.
+ *
+ * If you would like to distinguish just between the client and the compositor
+ * itself from the client's request, it can be done by getting the client
+ * credentials and by checking the PID of the client and the compositor's PID.
+ * Regarding the case in which the socketpair() is being used, you need to be
+ * careful. Please note the documentation for wl_client_get_credentials().
+ *
+ * This function can be used for a compositor to validate a request from
+ * a client if there are additional information provided from the client's
+ * file descriptor. For instance, suppose you can get the security contexts
+ * from the client's file descriptor. The compositor can validate the client's
+ * request with the contexts and make a decision whether it permits or deny it.
+ *
+ * \memberof wl_client
+ */
+WL_EXPORT int
+wl_client_get_fd(struct wl_client *client)
+{
+	return wl_connection_get_fd(client->connection);
+}
+
 /** Look up an object in the client name space
  *
  * \param client The client object
@@ -496,8 +533,8 @@ wl_client_get_credentials(struct wl_client *client,
  * \return The object or NULL if there is not object for the given ID
  *
  * This looks up an object in the client object name space by its
- * object ID.  
- * 
+ * object ID.
+ *
  * \memberof wl_client
  */
 WL_EXPORT struct wl_resource *
@@ -585,12 +622,12 @@ wl_resource_find_for_client(struct wl_list *list, struct wl_client *client)
 	if (client == NULL)
 		return NULL;
 
-        wl_list_for_each(resource, list, link) {
-                if (resource->client == client)
-                        return resource;
-        }
+	wl_list_for_each(resource, list, link) {
+		if (resource->client == client)
+			return resource;
+	}
 
-        return NULL;
+	return NULL;
 }
 
 WL_EXPORT struct wl_client *
@@ -672,7 +709,7 @@ wl_client_destroy(struct wl_client *client)
 	wl_map_for_each(&client->objects, destroy_resource, &serial);
 	wl_map_release(&client->objects);
 	wl_event_source_remove(client->source);
-	wl_connection_destroy(client->connection);
+	close(wl_connection_destroy(client->connection));
 	wl_list_remove(&client->link);
 	free(client);
 }
@@ -693,6 +730,11 @@ registry_bind(struct wl_client *client,
 		wl_resource_post_error(resource,
 				       WL_DISPLAY_ERROR_INVALID_OBJECT,
 				       "invalid global %s (%d)", interface, name);
+	else if (version == 0)
+		wl_resource_post_error(resource,
+				       WL_DISPLAY_ERROR_INVALID_OBJECT,
+				       "invalid version for global %s (%d): 0 is not a valid version",
+				       interface, name);
 	else if (global->version < version)
 		wl_resource_post_error(resource,
 				       WL_DISPLAY_ERROR_INVALID_OBJECT,
@@ -771,27 +813,25 @@ destroy_client_display_resource(struct wl_resource *resource)
 	resource->client->display_resource = NULL;
 }
 
-static void
-bind_display(struct wl_client *client,
-	     void *data, uint32_t version, uint32_t id)
+static int
+bind_display(struct wl_client *client, struct wl_display *display)
 {
-	struct wl_display *display = data;
-
 	client->display_resource =
-		wl_resource_create(client, &wl_display_interface, 1, id);
+		wl_resource_create(client, &wl_display_interface, 1, 1);
 	if (client->display_resource == NULL) {
-		wl_client_post_no_memory(client);
-		return;
+		/* DON'T send no-memory error to client - it has no
+		 * resource to which it could post the event */
+		return -1;
 	}
 
 	wl_resource_set_implementation(client->display_resource,
 				       &display_interface, display,
 				       destroy_client_display_resource);
+	return 0;
 }
 
 /** Create Wayland display object.
  *
- * \param None
  * \return The Wayland display object. Null if failed to create
  *
  * This creates the wl_display object.
@@ -830,16 +870,55 @@ wl_display_create(void)
 
 	wl_array_init(&display->additional_shm_formats);
 
-	if (!wl_global_create(display, &wl_display_interface, 1,
-			      display, bind_display)) {
-		wl_event_loop_destroy(display->loop);
-		free(display);
-		return NULL;
-	}
-
 	return display;
 }
 
+static void
+wl_socket_destroy(struct wl_socket *s)
+{
+	if (s->source)
+		wl_event_source_remove(s->source);
+	if (s->addr.sun_path[0])
+		unlink(s->addr.sun_path);
+	if (s->fd >= 0)
+		close(s->fd);
+	if (s->lock_addr[0])
+		unlink(s->lock_addr);
+	if (s->fd_lock >= 0)
+		close(s->fd_lock);
+
+	free(s);
+}
+
+static struct wl_socket *
+wl_socket_alloc(void)
+{
+	struct wl_socket *s;
+
+	s = zalloc(sizeof *s);
+	if (!s)
+		return NULL;
+
+	s->fd = -1;
+	s->fd_lock = -1;
+
+	return s;
+}
+
+/** Destroy Wayland display object.
+ *
+ * \param display The Wayland display object which should be destroyed.
+ * \return None.
+ *
+ * This function emits the wl_display destroy signal, releases
+ * all the sockets added to this display, free's all the globals associated
+ * with this display, free's memory of additional shared memory formats and
+ * destroy the display object.
+ *
+ * \sa wl_display_add_destroy_listener
+ *
+ * \memberof wl_display
+ */
 WL_EXPORT void
 wl_display_destroy(struct wl_display *display)
 {
@@ -849,12 +928,7 @@ wl_display_destroy(struct wl_display *display)
 	wl_signal_emit(&display->destroy_signal, display);
 
 	wl_list_for_each_safe(s, next, &display->socket_list, link) {
-		wl_event_source_remove(s->source);
-		unlink(s->addr.sun_path);
-		close(s->fd);
-		unlink(s->lock_addr);
-		close(s->fd_lock);
-		free(s);
+		wl_socket_destroy(s);
 	}
 	wl_event_loop_destroy(display->loop);
 
@@ -874,9 +948,17 @@ wl_global_create(struct wl_display *display,
 	struct wl_global *global;
 	struct wl_resource *resource;
 
-	if (interface->version < version) {
-		wl_log("wl_global_create: implemented version higher "
-		       "than interface version%m\n");
+	if (version < 1) {
+		wl_log("wl_global_create: failing to create interface "
+		       "'%s' with version %d because it is less than 1\n",
+			interface->name, version);
+		return NULL;
+	}
+
+	if (version > interface->version) {
+		wl_log("wl_global_create: implemented version for '%s' "
+		       "higher than interface version (%d > %d)\n",
+		       interface->name, version, interface->version);
 		return NULL;
 	}
 
@@ -921,7 +1003,7 @@ wl_global_destroy(struct wl_global *global)
  *
  * This function returns the most recent serial number, but does not
  * increment it.
- * 
+ *
  * \memberof wl_display
  */
 WL_EXPORT uint32_t
@@ -936,7 +1018,7 @@ wl_display_get_serial(struct wl_display *display)
  *
  * This function increments the display serial number and returns the
  * new value.
- * 
+ *
  * \memberof wl_display
  */
 WL_EXPORT uint32_t
@@ -1009,50 +1091,57 @@ socket_data(int fd, uint32_t mask, void *data)
 }
 
 static int
-get_socket_lock(struct wl_socket *socket)
+wl_socket_lock(struct wl_socket *socket)
 {
 	struct stat socket_stat;
-	int fd_lock;
 
 	snprintf(socket->lock_addr, sizeof socket->lock_addr,
 		 "%s%s", socket->addr.sun_path, LOCK_SUFFIX);
 
-	fd_lock = open(socket->lock_addr, O_CREAT | O_CLOEXEC,
-	               (S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP));
+	socket->fd_lock = open(socket->lock_addr, O_CREAT | O_CLOEXEC,
+			       (S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP));
 
-	if (fd_lock < 0) {
+	if (socket->fd_lock < 0) {
 		wl_log("unable to open lockfile %s check permissions\n",
 			socket->lock_addr);
-		return -1;
+		goto err;
 	}
 
-	if (flock(fd_lock, LOCK_EX | LOCK_NB) < 0) {
+	if (flock(socket->fd_lock, LOCK_EX | LOCK_NB) < 0) {
 		wl_log("unable to lock lockfile %s, maybe another compositor is running\n",
 			socket->lock_addr);
-		close(fd_lock);
-		return -1;
+		goto err_fd;
 	}
 
 	if (stat(socket->addr.sun_path, &socket_stat) < 0 ) {
 		if (errno != ENOENT) {
 			wl_log("did not manage to stat file %s\n",
 				socket->addr.sun_path);
-			close(fd_lock);
-			return -1;
+			goto err_fd;
 		}
 	} else if (socket_stat.st_mode & S_IWUSR ||
 		   socket_stat.st_mode & S_IWGRP) {
 		unlink(socket->addr.sun_path);
 	}
 
-	return fd_lock;
+	return 0;
+err_fd:
+	close(socket->fd_lock);
+	socket->fd_lock = -1;
+err:
+	*socket->lock_addr = 0;
+	/* we did not set this value here, but without lock the
+	 * socket won't be created anyway. This prevents the
+	 * wl_socket_destroy from unlinking already existing socket
+	 * created by other compositor */
+	*socket->addr.sun_path = 0;
+
+	return -1;
 }
 
-WL_EXPORT int
-wl_display_add_socket(struct wl_display *display, const char *name)
+static int
+wl_socket_init_for_display_name(struct wl_socket *s, const char *name)
 {
-	struct wl_socket *s;
-	socklen_t size;
 	int name_size;
 	const char *runtime_dir;
 
@@ -1066,62 +1155,44 @@ wl_display_add_socket(struct wl_display *display, const char *name)
 		return -1;
 	}
 
-	s = malloc(sizeof *s);
-	if (s == NULL)
-		return -1;
-
-	s->fd = wl_os_socket_cloexec(PF_LOCAL, SOCK_STREAM, 0);
-	if (s->fd < 0) {
-		free(s);
-		return -1;
-	}
-
-	if (name == NULL)
-		name = getenv("WAYLAND_DISPLAY");
-	if (name == NULL)
-		name = "wayland-0";
-
-	memset(&s->addr, 0, sizeof s->addr);
 	s->addr.sun_family = AF_LOCAL;
 	name_size = snprintf(s->addr.sun_path, sizeof s->addr.sun_path,
 			     "%s/%s", runtime_dir, name) + 1;
 
+	s->display_name = (s->addr.sun_path + name_size - 1) - strlen(name);
+
 	assert(name_size > 0);
 	if (name_size > (int)sizeof s->addr.sun_path) {
 		wl_log("error: socket path \"%s/%s\" plus null terminator"
 		       " exceeds 108 bytes\n", runtime_dir, name);
-		close(s->fd);
-		free(s);
+		*s->addr.sun_path = 0;
 		/* to prevent programs reporting
 		 * "failed to add socket: Success" */
 		errno = ENAMETOOLONG;
 		return -1;
-	};
+	}
 
-	s->fd_lock = get_socket_lock(s);
-	if (s->fd_lock < 0) {
-		close(s->fd);
-		free(s);
+	return 0;
+}
+
+static int
+_wl_display_add_socket(struct wl_display *display, struct wl_socket *s)
+{
+	socklen_t size;
+
+	s->fd = wl_os_socket_cloexec(PF_LOCAL, SOCK_STREAM, 0);
+	if (s->fd < 0) {
 		return -1;
 	}
 
-	size = offsetof (struct sockaddr_un, sun_path) + name_size;
+	size = offsetof (struct sockaddr_un, sun_path) + strlen(s->addr.sun_path);
 	if (bind(s->fd, (struct sockaddr *) &s->addr, size) < 0) {
 		wl_log("bind() failed with error: %m\n");
-		close(s->fd);
-		unlink(s->lock_addr);
-		close(s->fd_lock);
-		free(s);
 		return -1;
 	}
 
-	if (listen(s->fd, 1) < 0) {
+	if (listen(s->fd, 128) < 0) {
 		wl_log("listen() failed with error: %m\n");
-		unlink(s->addr.sun_path);
-		close(s->fd);
-		unlink(s->lock_addr);
-		close(s->fd_lock);
-		free(s);
 		return -1;
 	}
 
@@ -1129,14 +1200,148 @@ wl_display_add_socket(struct wl_display *display, const char *name)
 					 WL_EVENT_READABLE,
 					 socket_data, display);
 	if (s->source == NULL) {
-		unlink(s->addr.sun_path);
-		close(s->fd);
-		unlink(s->lock_addr);
-		close(s->fd_lock);
-		free(s);
 		return -1;
 	}
+
 	wl_list_insert(display->socket_list.prev, &s->link);
+	return 0;
+}
+
+WL_EXPORT const char *
+wl_display_add_socket_auto(struct wl_display *display)
+{
+	struct wl_socket *s;
+	int displayno = 0;
+	char display_name[16] = "";
+
+	/* A reasonable number of maximum default sockets. If
+	 * you need more than this, use the explicit add_socket API. */
+	const int MAX_DISPLAYNO = 32;
+
+	s = wl_socket_alloc();
+	if (s == NULL)
+		return NULL;
+
+	do {
+		snprintf(display_name, sizeof display_name, "wayland-%d", displayno);
+		if (wl_socket_init_for_display_name(s, display_name) < 0) {
+			wl_socket_destroy(s);
+			return NULL;
+		}
+
+		if (wl_socket_lock(s) < 0)
+			continue;
+
+		if (_wl_display_add_socket(display, s) < 0) {
+			wl_socket_destroy(s);
+			return NULL;
+		}
+
+		return s->display_name;
+	} while (displayno++ < MAX_DISPLAYNO);
+
+	/* Ran out of display names. */
+	wl_socket_destroy(s);
+	errno = EINVAL;
+	return NULL;
+}
+
+/**  Add a socket with an existing fd to Wayland display for the clients to connect.
+ *
+ * \param display Wayland display to which the socket should be added.
+ * \param sock_fd The existing socket file descriptor to be used
+ * \return 0 if success. -1 if failed.
+ *
+ * The existing socket fd must already be created, opened, and locked.
+ * The fd must be properly set to CLOEXEC and bound to a socket file
+ * with both bind() and listen() already called.
+ *
+ * \memberof wl_display
+ */
+WL_EXPORT int
+wl_display_add_socket_fd(struct wl_display *display, int sock_fd)
+{
+	struct wl_socket *s;
+	struct stat buf;
+
+	/* Require a valid fd or fail */
+	if (sock_fd < 0 || fstat(sock_fd, &buf) < 0 || !S_ISSOCK(buf.st_mode)) {
+		return -1;
+	}
+
+	s = wl_socket_alloc();
+	if (s == NULL)
+		return -1;
+
+	s->source = wl_event_loop_add_fd(display->loop, sock_fd,
+					 WL_EVENT_READABLE,
+					 socket_data, display);
+	if (s->source == NULL) {
+		wl_log("failed to establish event source\n");
+		wl_socket_destroy(s);
+		return -1;
+	}
+
+	/* Reuse the existing fd */
+	s->fd = sock_fd;
+
+	wl_list_insert(display->socket_list.prev, &s->link);
+
+	return 0;
+}
+
+/** Add a socket to Wayland display for the clients to connect.
+ *
+ * \param display Wayland display to which the socket should be added.
+ * \param name Name of the Unix socket.
+ * \return 0 if success. -1 if failed.
+ *
+ * This adds a Unix socket to Wayland display which can be used by clients to
+ * connect to Wayland display.
+ *
+ * If NULL is passed as name, then it would look for WAYLAND_DISPLAY env
+ * variable for the socket name. If WAYLAND_DISPLAY is not set, then default
+ * wayland-0 is used.
+ *
+ * The Unix socket will be created in the directory pointed to by environment
+ * variable XDG_RUNTIME_DIR. If XDG_RUNTIME_DIR is not set, then this function
+ * fails and returns -1.
+ *
+ * The length of socket path, i.e., the path set in XDG_RUNTIME_DIR and the
+ * socket name, must not exceed the maxium length of a Unix socket path.
+ * The function also fails if the user do not have write permission in the
+ * XDG_RUNTIME_DIR path or if the socket name is already in use.
+ *
+ * \memberof wl_display
+ */
+WL_EXPORT int
+wl_display_add_socket(struct wl_display *display, const char *name)
+{
+	struct wl_socket *s;
+
+	s = wl_socket_alloc();
+	if (s == NULL)
+		return -1;
+
+	if (name == NULL)
+		name = getenv("WAYLAND_DISPLAY");
+	if (name == NULL)
+		name = "wayland-0";
+
+	if (wl_socket_init_for_display_name(s, name) < 0) {
+		wl_socket_destroy(s);
+		return -1;
+	}
+
+	if (wl_socket_lock(s) < 0) {
+		wl_socket_destroy(s);
+		return -1;
+	}
+
+	if (_wl_display_add_socket(display, s) < 0) {
+		wl_socket_destroy(s);
+		return -1;
+	}
 
 	return 0;
 }
@@ -1204,11 +1409,10 @@ wl_resource_create(struct wl_client *client,
 	resource->version = version;
 	resource->dispatcher = NULL;
 
-	if (wl_map_insert_at(&client->objects, 0, resource->object.id, resource) < 0) {
+	if (wl_map_insert_at(&client->objects, 0, id, resource) < 0) {
 		wl_resource_post_error(client->display_resource,
 				       WL_DISPLAY_ERROR_INVALID_OBJECT,
-				       "invalid new id %d",
-				       resource->object.id);
+				       "invalid new id %d", id);
 		free(resource);
 		return NULL;
 	}
@@ -1222,7 +1426,59 @@ wl_log_set_handler_server(wl_log_func_t handler)
 	wl_log_handler = handler;
 }
 
-/* Deprecated functions below. */
+/** Add support for a wl_shm pixel format
+ *
+ * \param display The display object
+ * \param format The wl_shm pixel format to advertise
+ * \return A pointer to the wl_shm format that was added to the list
+ * or NULL if adding it to the list failed.
+ *
+ * Add the specified wl_shm format to the list of formats the wl_shm
+ * object advertises when a client binds to it.  Adding a format to
+ * the list means that clients will know that the compositor supports
+ * this format and may use it for creating wl_shm buffers.  The
+ * compositor must be able to handle the pixel format when a client
+ * requests it.
+ *
+ * The compositor by default supports WL_SHM_FORMAT_ARGB8888 and
+ * WL_SHM_FORMAT_XRGB8888.
+ *
+ * \memberof wl_display
+ */
+WL_EXPORT uint32_t *
+wl_display_add_shm_format(struct wl_display *display, uint32_t format)
+{
+	uint32_t *p = NULL;
+
+	p = wl_array_add(&display->additional_shm_formats, sizeof *p);
+
+	if (p != NULL)
+		*p = format;
+	return p;
+}
+
+/**
+ * Get list of additional wl_shm pixel formats
+ *
+ * \param display The display object
+ *
+ * This function returns the list of addition wl_shm pixel formats
+ * that the compositor supports.  WL_SHM_FORMAT_ARGB8888 and
+ * WL_SHM_FORMAT_XRGB8888 are always supported and not included in the
+ * array, but all formats added through wl_display_add_shm_format()
+ * will be in the array.
+ *
+ * \sa wl_display_add_shm_format()
+ *
+ * \memberof wl_display
+ */
+struct wl_array *
+wl_display_get_additional_shm_formats(struct wl_display *display)
+{
+	return &display->additional_shm_formats;
+}
+
+/** \cond */ /* Deprecated functions below. */
 
 uint32_t
 wl_client_add_resource(struct wl_client *client,
@@ -1319,54 +1575,9 @@ wl_display_remove_global(struct wl_display *display, struct wl_global *global)
 	wl_global_destroy(global);
 }
 
-/** Add support for a wl_shm pixel format
- *
- * \param display The display object
- * \param format The wl_shm pixel format to advertise
- * \return A pointer to the wl_shm format that was added to the list
- * or NULL if adding it to the list failed.
- *
- * Add the specified wl_shm format to the list of formats the wl_shm
- * object advertises when a client binds to it.  Adding a format to
- * the list means that clients will know that the compositor supports
- * this format and may use it for creating wl_shm buffers.  The
- * compositor must be able to handle the pixel format when a client
- * requests it.
- *
- * The compositor by default supports WL_SHM_FORMAT_ARGB8888 and
- * WL_SHM_FORMAT_XRGB8888.
- *
- * \memberof wl_display
+/** \endcond */
+
+/* Functions at the end of this file are deprecated.  Instead of adding new
+ * code here, add it before the comment above that states:
+ * Deprecated functions below.
  */
-WL_EXPORT uint32_t *
-wl_display_add_shm_format(struct wl_display *display, uint32_t format)
-{
-	uint32_t *p = NULL;
-
-	p = wl_array_add(&display->additional_shm_formats, sizeof *p);
-
-	if (p != NULL)
-		*p = format;
-	return p;
-}
-
-/**
- * Get list of additional wl_shm pixel formats
- *
- * \param display The display object
- *
- * This function returns the list of addition wl_shm pixel formats
- * that the compositor supports.  WL_SHM_FORMAT_ARGB8888 and
- * WL_SHM_FORMAT_XRGB8888 are always supported and not included in the
- * array, but all formats added through wl_display_add_shm_format()
- * will be in the array.
- * 
- * \sa wl_display_add_shm_format()
- * 
- * \memberof wl_display
- */
-struct wl_array *
-wl_display_get_additional_shm_formats(struct wl_display *display)
-{
-	return &display->additional_shm_formats;
-}
